from collections.abc import Callable
from typing import Any

import jax

type BooleanNumeric = Any


def while_loop[T](
    cond_fun: Callable[[T], BooleanNumeric],
    body_fun: Callable[[T], T],
    init_val: T,
    *,
    jit: bool = True,
) -> T:
    if jit:
        return jax.lax.while_loop(cond_fun, body_fun, init_val)
    val: T = init_val
    while cond_fun(val):
        val = body_fun(val)
    return val



from ._define import PyTreeType, define, frozen, frozen_static
from ._field_specifiers import FieldType, array, auto, field, static
from ._register import register_fieldz

__all__ = [
    "FieldType",
    "PyTreeType",
    "array",
    "auto",
    "define",
    "field",
    "frozen",
    "frozen_static",
    "register_fieldz",
    "static",
]



from collections.abc import Callable, Sequence
from typing import Any

import fieldz

from jarp.tree._filters import is_data
from jarp.tree.codegen import register_generic

from ._field_specifiers import FieldType


def register_fieldz[T: type](
    cls: T,
    data_fields: Sequence[str] | None = None,
    meta_fields: Sequence[str] | None = None,
    auto_fields: Sequence[str] | None = None,
    *,
    filter_spec: Callable[[Any], bool] = is_data,
    bypass_setattr: bool | None = None,
) -> T:
    if data_fields is None:
        data_fields = _filter_field_names(cls, FieldType.DATA)
    if meta_fields is None:
        meta_fields = _filter_field_names(cls, FieldType.META)
    if auto_fields is None:
        auto_fields = _filter_field_names(cls, FieldType.AUTO)
    register_generic(
        cls,
        data_fields,
        meta_fields,
        auto_fields,
        filter_spec=filter_spec,
        bypass_setattr=bypass_setattr,
    )
    return cls


def _filter_field_names(cls: type, field_type: FieldType) -> list[str]:
    return [
        f.name
        for f in fieldz.fields(cls)
        if FieldType(f.metadata.get("static")) is field_type
    ]



from collections.abc import Iterable
from typing import Any, cast

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, ArrayLike, DTypeLike, Shaped

from ._filters import combine_leaves, partition_leaves
from .attrs import frozen_static

type PyTreeDef = Any
type Shape = tuple[int, ...]
type Vector = Shaped[Array, " N"]


@frozen_static
class Structure[T]:
    dtype: DTypeLike
    meta_leaves: tuple[Any, ...]
    offsets: tuple[int, ...]
    shapes: tuple[Shape | None, ...]
    treedef: PyTreeDef

    @property
    def is_leaf(self) -> bool:
        return jtu.treedef_is_leaf(self.treedef)

    def ravel(self, tree: T | Array) -> Vector:
        if isinstance(tree, Array):
            # do not flatten if already flat
            return jnp.ravel(tree)
        leaves: list[Any]
        treedef: PyTreeDef
        leaves, treedef = jax.tree.flatten(tree)
        assert treedef == self.treedef
        data_leaves: list[Array | None]
        meta_leaves: list[Any]
        data_leaves, meta_leaves = partition_leaves(leaves)
        assert meta_leaves == self.meta_leaves
        return _ravel(data_leaves)

    def unravel(self, flat: T | ArrayLike, dtype: DTypeLike | None = None) -> T:
        if not isinstance(flat, Array):
            # do not unravel if already a pytree
            assert jax.tree.structure(flat) == self.treedef
            return cast("T", flat)
        flat: Array = jnp.asarray(flat, self.dtype if dtype is None else dtype)
        if self.is_leaf:
            return cast("T", jnp.reshape(flat, self.shapes[0]))
        data_leaves: list[Array | None] = _unravel(flat, self.offsets, self.shapes)
        leaves: list[Any] = combine_leaves(data_leaves, self.meta_leaves)
        return jax.tree.unflatten(self.treedef, leaves)


def ravel[T](tree: T) -> tuple[Array, Structure[T]]:
    leaves: list[Any]
    treedef: PyTreeDef
    leaves, treedef = jax.tree.flatten(tree)
    dynamic_leaves: list[Any | None]
    static_leaves: list[Any | None]
    dynamic_leaves, static_leaves = partition_leaves(leaves)
    flat: Array = _ravel(dynamic_leaves)
    structure: Structure[T] = Structure(
        offsets=_offsets_from_leaves(dynamic_leaves),
        shapes=_shapes_from_leaves(dynamic_leaves),
        meta_leaves=tuple(static_leaves),
        treedef=treedef,
        dtype=flat.dtype,
    )
    return flat, structure


def _offsets_from_leaves(leaves: Iterable[Any | None]) -> tuple[int, ...]:
    offsets: list[int] = []
    i: int = 0
    for leaf in leaves:
        if leaf is not None:
            i += jnp.size(leaf)
        offsets.append(i)
    return tuple(offsets)


@jax.jit
def _ravel(leaves: Iterable[Array | None]) -> Array:
    return jnp.concatenate([leaf for leaf in leaves if leaf is not None], axis=None)


def _shapes_from_leaves(leaves: Iterable[Any | None]) -> tuple[Shape | None, ...]:
    return tuple(None if leaf is None else jnp.shape(leaf) for leaf in leaves)


@jax.jit(static_argnums=(1, 2))
def _unravel(
    flat: Array, offsets: tuple[int, ...], shapes: tuple[Shape | None, ...]
) -> list[Array | None]:
    assert jnp.size(flat) == offsets[-1]
    chunks: list[Array] = jnp.split(flat, offsets[:-1])
    leaves: list[Array | None] = []
    for chunk, shape in zip(chunks, shapes, strict=True):
        if shape is None:
            assert jnp.size(chunk) == 0
            leaves.append(None)
        else:
            leaves.append(jnp.reshape(chunk, shape))
    return leaves



from __future__ import annotations

import enum
from collections.abc import Callable, Mapping
from typing import Any, TypedDict, Unpack

import attrs
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike
from liblaf import grapes


class FieldType(enum.StrEnum):
    AUTO = enum.auto()
    DATA = enum.auto()
    META = enum.auto()

    @classmethod
    def _missing_(cls, value: object) -> Any:
        if isinstance(value, str):
            return cls(value.lower())
        if value is True:
            return cls.META
        if value is False or value is None:
            return cls.DATA
        return None

    def __bool__(self) -> bool:
        match self:
            case FieldType.META:
                return True
            case FieldType.AUTO | FieldType.DATA:
                # for consistency with `jax.tree_util.register_dataclass`
                return False


class FieldOptions[T](TypedDict, total=False):
    default: T
    validator: attrs._ValidatorArgType[T] | None
    repr: attrs._ReprArgType
    hash: bool | None
    init: bool
    metadata: Mapping[Any, Any] | None
    converter: (
        attrs._ConverterType
        | list[attrs._ConverterType]
        | tuple[attrs._ConverterType, ...]
        | None
    )
    factory: Callable[[], T] | None
    kw_only: bool | None
    eq: attrs._EqOrderType | None
    order: attrs._EqOrderType | None
    on_setattr: attrs._OnSetAttrArgType | None
    alias: str | None
    type: type | None

    static: FieldType | bool | None


def array(**kwargs: Unpack[FieldOptions[ArrayLike | None]]) -> Array:
    if "default" in kwargs and "factory" not in kwargs:
        default: ArrayLike | None = kwargs.pop("default")  # pyright: ignore[reportAssignmentType]
        if default is not None:
            default = jnp.asarray(default)
        kwargs["factory"] = lambda: default
    return field(**kwargs)  # pyright: ignore[reportArgumentType, reportCallIssue]


@grapes.wraps(attrs.field)
def auto(**kwargs) -> Any:
    kwargs.setdefault("static", FieldType.AUTO)
    return field(**kwargs)


@grapes.wraps(attrs.field)
def field(**kwargs) -> Any:
    if "static" in kwargs:
        kwargs["metadata"] = {
            "static": kwargs.pop("static"),
            **(kwargs.get("metadata") or {}),
        }
    return attrs.field(**kwargs)


@grapes.wraps(attrs.field)
def static(**kwargs) -> Any:
    # for consistency with `jax.tree_util.register_dataclass`
    kwargs.setdefault("static", True)
    return field(**kwargs)
